;;;;;;;;;;;;;;;;;;;;;;;;;;;;;<--OGI-->;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                        ;;
;;             Center for Spoken Language Understanding                   ;;
;;        Oregon Graduate Institute of Science & Technology               ;;
;;                         Portland, OR USA                               ;;
;;                        Copyright (c) 1999                              ;;
;;                                                                        ;;
;;      This module is not part of the CSTR/University of Edinburgh       ;;
;;               release of the Festival TTS system.                      ;;
;;                                                                        ;;
;;  In addition to any conditions disclaimers below, please see the file  ;;
;;  "license_ogi_tts.txt" distributed with this software for information  ;;
;;  on usage and redistribution, and for a DISCLAIMER OF ALL WARRANTIES.  ;;
;;                                                                        ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;<--OGI-->;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;  Festival support for SABLE 0.2 the SGML/XML based mark up    ;;
;;;  language.                                                            ;;
;;;                                                                       ;;
;;;  This is XML version requiring Edinburgh's LTG's rxp XML parser as    ;;
;;;  distributed with Festival                                            ;;
;;;                                                                       ;;

(require_module 'rxp)
(require 'ogi_languages)

 ;; Remember where to find these two XML entities.                        ;;
(xml_register_id "-//SABLE//DTD SABLE speech mark up//EN"
    (path-append libdir "ogi_Sable.v0_2.dtd")
    )

(xml_register_id "-//SABLE//ENTITIES Added Latin 1 for SABLE//EN"
     (path-append libdir  "sable-latin.ent")
     )

(defvar SABLE_RXDOUBLE "-?\\(\\([0-9]+\\.[0-9]*\\)\\|\\([0-9]+\\)\\|\\(\\.[0-9]+\\)\\)\\([eE][---+]?[0-9]+\\)?")

(defvar ogisable_pitch_base_map
  '((highest 1.2)
    (high 1.1)
    (medium 1.0)
    (default 1.0)
    (low 0.9)
    (lowest 0.8)))
(defvar ogisable_pitch_med_map
  '((highest 1.2)
    (high 1.1)
    (medium 1.0)
    (default 1.0)
    (low 0.9)
    (lowest 0.8)))
(defvar ogisable_pitch_range_map
  '((largest 1.2)
    (large 1.1)
    (medium 1.0)
    (default 1.0)
    (small 0.9)
    (smallest 0.8)))
(defvar ogisable_rate_speed_map
  '((fastest 1.5)
    (fast 1.2)
    (medium 1.0)
    (default 1.0)
    (slow 0.8)
    (slowest 0.6)))
(defvar ogisable_volume_level_map
  '((loudest 2.0) 
    (loud 1.5)
    (default 1.0)
    (medium 1.0)
    (quiet 0.5)))

(define (ogisable_init_globals)
  (set! utts nil)
  (set! ogisable_omitted_mode nil)
  (set! ogisable_word_features_stack nil)
  (set! ogisable_pitch_context nil)
  (set! ogisable_vol_context nil)
  (set! ogisable_vol_type 'no_change)
  (set! ogisable_vol_factor 1.0)
  (set! ogisable_pronID 0)
  (set! ogisable_pitchID 0)
  (set! ogisable_volumeID 0)
  (set! ogisable_vqualityID 0)
  (set! ogisable_current_language 'americanenglish)
  (set! ogisable_unsupported_language nil)
  (set! ogisable_language_stack nil)
  ;;(set! ogisable_current_speaker 'voice_mwm_diphone)
  (set! ogisable_speaker_stack nil)
)

(define (ogisable_token_to_words token name)
  "(ogisable_token_to_words utt token name)
OGISABLE mode token specific analysis."
  ;(print (item.features token))
  (cond
   ((or ogisable_omitted_mode ogisable_unsupported_language)
      ;; don't say anything (whole utterance)
      nil)
   ((and (not (equal? (item.feat token "ogisable_sub") 0))
   (not (string-equal (item.feat token "ogisable_pronID") (item.feat token "p.ogisable_pronID"))))
      ;; identical ogisable_sub is attached to all tokens between <pron> and </pron>
      ;; take ogisable_sub from first token only
      (read-from-string (item.feat token "ogisable_sub")))
   ((and (not (equal? (item.feat token "ogisable_ipa") 0))
   (not (string-equal (item.feat token "ogisable_pronID") (item.feat token "p.ogisable_pronID"))))
      (item.set_feat token "punc" "")
      (item.set_feat token "phonemes" 
        (format nil "%l" (ogisable-map-ipa 
        (read-from-string (item.feat token "ogisable_ipa")))))
      (list name))
   ((and (not (equal? (item.feat token "ogisable_segdur") 0))
   (not (string-equal (item.feat token "ogisable_pronID") (item.feat token "p.ogisable_pronID"))))
      (let ((segdurs (ogisable-map-segdur (item.feat token "ogisable_segdur"))))
      (item.set_feat token "phonemes" (format nil "%l" (car segdurs)))
      (item.set_feat token "punc" "")
      (item.set_feat token "durations" (format nil "%l" (cadr segdurs)))
      (list name)))
   ;;((string-equal "1" (item.feat token "ogisable_ipa"))
      ;;  each token is an IPA phone
      ;;(item.set_feat token "phonemes" (format nil "%l" (ogisable-map-ipa name)))
      ;;(list name))
   ((string-equal "1" (item.feat token "ogisable_ignore"))
      ;; don't say anything (individual word)
      nil)
   ((string-equal "1" (item.feat token "ogisable_literal"))
      ;; Only deal with spell here
      (let ((subwords) (subword) (phons) (stretch 1.0) (newstretch) 
            (spellmode (item.feat token "spell_mode")))
        (item.set_feat token "pos" token.letter_pos)
        (if (string-equal spellmode "isolated") (begin
          (item.set_feat token "EMPH" "1")
          ;; put break before, after every token
          (if (item.prev token) (item.set_feat (item.prev token) "pbreak" "B"))
          (item.set_feat token "pbreak" "B")
          ;; spell first letter and create new token with remainder of string
          (set! subword (car (ogisable_previous_token_to_words token (car (symbolexplode name)))))
          (if (> (string-length name) 1) (begin
            (item.insert token (list (substring name 1 (string-length name))))
            (item.set_feat (item.next token) "ogisable_literal" "1")
            (item.set_feat (item.next token) "prepunctuation" "")
            (item.set_feat (item.next token) "spell_mode" "isolated")))
          ;; create rhythm by stretching duration depending on number of phonemes
          (if (consp subword) (set! subword (cadr (car subword))))
          (set! phons (find_phonemes subword token.letter_pos))
          (set! phons (cons 'pau phons))
          (item.set_feat token "phonemes" (format nil "%l" phons))
          ;(item.set_feat token "punc" "")
          (set! newstretch (/ 4 (length phons)))
          (if (string-equal subword "w") (set! newstretch (* 2 newstretch)))
          (if (equal? (item.feat token "dur_stretch") 0) 
            (item.set_feat token "dur_stretch" newstretch)
            ;; else
            (begin
              (set! stretch (item.feat token "dur_stretch"))
              (item.set_feat token "dur_stretch" (* newstretch (parse-number stretch)))))
          (set! subwords (list subword)))
          ;; else (inherited from earlier implementation .. length(name) = 1)
            (mapcar
             (lambda (letter)
                 ;; might be symbol or digit
                 (set! subword (ogisable_previous_token_to_words token letter))     
                   (if subwords
                     (set! subwords (append subwords subword))
                     ;; else
                     (set! subwords subword)))
               (symbolexplode name)))
        subwords))
   ((string-equal "1" (item.feat token "ogisable_syllabify"))
      (let ((subwords) (subword) (stretch 1.0) (pron) (phons) (syls) (syl) (token2)
            (syll_info (item.feat token "syllabify_info")))
        ;;(item.set_feat token "pos" token.letter_pos)
        (item.set_feat token "EMPH" "1")
        (item.set_feat token "no_postlex" "1")
        (if (equal? (item.feat token "dur_stretch") 0) 
          (item.set_feat token "dur_stretch" 2.0)
          ;; else
          (begin
            (set! stretch (item.feat token "dur_stretch"))
            (item.set_feat token "dur_stretch" (* 2.0 (parse-number stretch)))))
        ;; put break before, after every token
        (if (item.prev token) (item.set_feat (item.prev token) "pbreak" "B"))
        (item.set_feat token "pbreak" "B")
        ;; do automatic syllabification if syll_info not provided
        (if (string-equal "nil" syll_info) (begin
          (set! subwords (ogisable_previous_token_to_words token name))
          (set! subword (car subwords))
          ;; create new tokens for extra words from this token (e.g. '200' -> 'two' 'hundred') 
          (set! subwords (reverse (cdr subwords)))
          (while subwords (begin
            (item.insert token (list (car subwords)))
            (item.set_feat (item.next token) "ogisable_syllabify" "1")
            (item.set_feat (item.next token) "prepunctuation" "")
            (item.set_feat (item.next token) "syllabify_info" "0")
            (item.set_feat (item.next token) "dur_stretch" stretch)
            (item.set_feat (item.next token) "EMPH" "1")
            (set! subwords (cdr subwords))))
          ;; lookup pronunciation (no pos info available, which is too bad)
          (set! phons (find_phonemes subword nil))
          ;; find syllabic boundaries according to own heuristic (lex.syllabify does really bad job at the moment)
          (set! syls (syllabify phons)))
        ;; else: syll_info should be stringlist of symbols and .
          (begin 
            (set! syls (make-syl-list syll_info))
            ;; if this is first token in tag, make syll_info not repeated for other tokens
            ;; this fails when last token before tag was also syllabified
            (if (or (not (item.prev token)) 
                (not (string-equal (item.feat (item.prev token) "ogisable_syllabify") "1"))) 
              (begin
                (set! token2 (item.next token))
                (while (and token2 
                    (string-equal syll_info (item.feat token2 "syllabify_info"))) 
                  (begin
                  (item.set_feat token2 "ogisable_ignore" "1")
                  (set! token2 (item.next token2))))
             ))
          )
        )
        ;; use phonemes feature of this token for 1st syl and generate new tokens for other syls
        (set! syl (cons 'pau (car syls)))
        (item.set_feat token "phonemes" (format nil "%l" syl))
        (item.set_feat token "punc" "")
        (item.set_feat token "dur_stretch" (/ (* 3 (item.feat token "dur_stretch")) (length syl)))
        (set! syls (reverse (cdr syls))) 
        (while syls (begin
          (item.insert token (list subword))
          (item.set_feat (item.next token) "ogisable_syllabify" "1")
          (set! phons "")
          (set! syl (car syls))
          (while syl (begin 
            (set! phons (string-append phons (car syl) " "))
            (set! syl (cdr syl))))
          (item.set_feat (item.next token) "syllabify_info" (format nil "%s" phons))  
          (item.set_feat (item.next token) "prepunctuation" "")
          (item.set_feat (item.next token) "dur_stretch" stretch)
          (item.set_feat (item.next token) "EMPH" "1")
          (set! syls (cdr syls))))
        (list subword)
      ))
   ((not (equal? 0 (item.feat token "token_pos")))
      ;; bypass the prediction stage, if English
      ;(if (member_string (car (cdr (assoc 'Language Parameter)))
      ;       '(britishenglish americanenglish))
      ;(builtin_english_token_to_words token name)
      (ogisable_previous_token_to_words token name))
   ;; could be others here later
   (t  
      (ogisable_previous_token_to_words token name))))
      
(define (ogisable_Intonation utt)
  "(ogisable_Intonation utt)
   Redefinition of Intonation to include hook to modify Tobi labels. First Intonation 
   is predicted according to the method set in Int_Method. If Int_method is set to ToBI, the 
   Tobi labels are modified according to the token.tobi feature. Since Tobi labels should be
   specified on the syllable level, a single token.tobi label will be placed on the syllable
   with primary stress. If a ; separated sequence of labels is provided, one will be placed 
   on each syllable."
  (ogisable_previous_Intonation utt)
  (if (equal? Intonation_Tree (Param.get 'Int_Method)) (begin 
    (let ((token (utt.relation.first utt 'Token))
          (emph) (tobi) (words) (syls) (syl) (tmpsyl) (tmpsyls) (lab))
    (while token (begin
      (set! emph (item.feat token 'EMPH))
      (if (not (equal? emph 0)) 
        (set! tobi "H*")
        (set! tobi (item.feat token 'ogisable_tobi)))
      (if (not (equal? tobi 0)) (begin
        (set! words (item.daughters token))
        (while words
          (set! syls (item.relation.daughters (car words) "SylStructure"))
          (if (string-matches tobi ".*;.*")
            (while syls
              (set! lab (string-before tobi ";"))
              (set! tobi (string-after tobi ";"))
              (set! syl (car syls))
              (mapcar (lambda (a) (item.delete a)) 
                (item.relation.daughters syl "Intonation"))
              (if (not (item.relation syl "Intonation")) (begin
                (set! tmpsyl (item.prev syl))
                (while (and tmpsyl (not (item.relation tmpsyl "Intonation")))
                  (set! tmpsyl (item.prev tmpsyl)))
                (if tmpsyl  
                  (item.relation.insert tmpsyl "Intonation" syl 'after)
                  (utt.relation.append utt "Intonation" syl))))
              (item.relation.append_daughter syl "Intonation" (list lab))
              (set! syls (cdr syls)))
            ;; else find stressed syllable
            (begin
            (set! tmpsyls syls)
            (while tmpsyls  
              (if (equal? (item.feat (car tmpsyls) "stress") 1)
                (set! syls tmpsyls))
              (mapcar (lambda (a) (item.delete a)) 
                (item.relation.daughters (car tmpsyls) "Intonation"))
              (set! tmpsyls (cdr tmpsyls)))
            (if syls (begin
              (set! syl (car syls))
              (if (not (item.relation syl "Intonation")) (begin
                (set! tmpsyl (item.prev syl))
                (while (and tmpsyl (not (item.relation tmpsyl "Intonation")))
                  (set! tmpsyl (item.prev tmpsyl)))
                (if tmpsyl  
                  (item.relation.insert tmpsyl "Intonation" syl 'after)
                  (utt.relation.append utt "Intonation" syl))))
              (item.relation.append_daughter syl "Intonation" (list tobi)))
             )))
          (set! words (cdr words)))))
      (set! token (item.next token)))))))
  utt
)
   
(defvar ogisable_elements
'(
  ("(SABLE" (ATTLIST UTT)
    (eval (list ogisable_current_speaker))  ;; so we know what state we start in
    (ogisable_setup_voice_params)
    (init_ogisable_token_to_words)
    nil
  )
  (")SABLE" (ATTLIST UTT)
    (xxml_synth UTT)  ;;  Synthesize the remaining tokens
    nil
  )
  ;; The following commands can cause utterance breaks.
  ;; They can change some global Festival settings, like the voice or the avg pitch
  ;; in such case, a stack is kept of previous settings, which are restored
  ;; after the closing tag 
  ;; For operation with CSLU toolkit, appropriate pre-chunking probably necessary.
  
  ;; the language tags are in fact just shortcuts to specific voices.
  ;; this is defined in languages.scm
  ("(LANGUAGE" (ATTLIST UTT)
   (xxml_synth UTT)
   (set! ogisable_language_stack 
   (cons 
    (list ogisable_current_language ogisable_unsupported_language)
    ogisable_language_stack))
   ;; Select a new language
   (let ((language (upcase (car (xxml_attval "ID" ATTLIST)))))
     (cond
      ((or (string-equal language "SPANISH")
     (string-equal language "ES"))
       (set! ogisable_current_language 'mexicanspanish)
       (set! ogisable_unsupported_language nil)
       (select_language 'mexicanspanish))
      ((or (string-equal language "ENGLISH")
     (string-equal language "EN"))
       (set! ogisable_current_language 'americanenglish)
       (set! ogisable_unsupported_language nil)
       (select_language 'americanenglish))
      ((or (string-equal language "GERMAN")
     (string-equal language "DE"))
       (set! ogisable_current_language 'german)
       (set! ogisable_unsupported_language nil)
       (select_language 'german))
      (t  ;; skip languages you don't know
       ;; BUG: if current language isn't English this wont work
       (apply_hooks tts_hooks
        (eval (list 'Utterance 'Text
        (string-append "Some text in " language))))
       (set! ogisable_unsupported_language t)))
     nil))
  (")LANGUAGE" (ATTLIST UTT)
   (xxml_synth UTT)
   (set! ogisable_unsupported_language (car (cdr (car ogisable_language_stack))))
   (set! ogisable_current_language (car (car ogisable_language_stack)))
   (set! ogisable_language_stack (cdr ogisable_language_stack))
   (if (not ogisable_omitted_mode)
       (begin
   (select_language ogisable_current_language)
   (ogisable_setup_voice_params)))
   nil)
  ("(SPEAKER" (ATTLIST UTT)
   ;; Status: GENDER/AGE ignored, should be done by ogisable-def-speaker 
   ;;         function to define Festival voices to SABLE
   (xxml_synth UTT)
   (set! ogisable_speaker_stack (cons ogisable_current_speaker ogisable_speaker_stack))
   (let ((speaker (car (xxml_attval "NAME" ATTLIST))))
     (cond
      ((not equal? ogisable_current_language 'americanenglish)
       (print "SABLE: chosen unknown voice, current voice unchanged"))
      ((or (equal? speaker 'male1) (equal? speaker 'mwm))
       (set! ogisable_current_speaker 'voice_mwm_diphone)
       (voice_mwm_diphone))
      ((or (equal? speaker 'male2) (equal? speaker 'jph))
       (set! ogisable_current_speaker 'voice_jph_diphone)
       (voice_jph_diphone))
      ((or (equal? speaker 'male3) (equal? speaker 'aec))
       (set! ogisable_current_speaker 'voice_aec_diphone)
       (voice_aec_diphone))
      ((or (equal? speaker 'female) (equal? speaker 'tll))
       (set! ogisable_current_speaker 'voice_tll_diphone)
       (voice_tll_diphone))
      ((or (equal? speaker 'spanish) (equal? speaker 'abc))
       (set! ogisable_current_speaker 'voice_abc_diphone)
       (voice_abc_diphone))
      ((or (equal? speaker 'german) (equal? speaker 'axk))
       (set! ogisable_current_speaker 'voice_axk_diphone)
       (voice_axk_diphone))
      (t
       (print "SABLE: selecting unknown voice")
       (set! ogisable_current_speaker voice_mwm_diphone)
       (voice_mwm_diphone))))
    (ogisable_setup_voice_params)
    (init_ogisable_token_to_words)
   nil)
  (")SPEAKER" (ATTLIST UTT)
   (xxml_synth UTT)
   (set! ogisable_current_speaker (car ogisable_speaker_stack))
   (set! ogisable_speaker_stack (cdr ogisable_speaker_stack))
   (eval (list ogisable_current_speaker))
   (ogisable_setup_voice_params)
   (init_ogisable_token_to_words)
   nil)
  ("BREAK" (ATTLIST UTT)
   ;; may cause an utterance break
   (let ((level (upcase (car (xxml_attval "LEVEL" ATTLIST))))
         (msec (upcase (car (xxml_attval "MSEC" ATTLIST)))))
   (cond
     ((not (equal? msec "NIL"))
       ;; insert silence and set duration
       (set! msec (/ (parse-number msec) 100)) ;; transform to stretch, assuming orig 100 ms
       (let ((last_token (utt.relation.last UTT 'Token)))
       ;(if last_token 
        ; (item.insert last_token '(BREAK nil))
        ; (utt.relation.append UTT 'Token '(BREAK nil)))
       (utt.relation.append UTT 'Token '(BREAK ((prepunctuation ""))))
       (let ((it (utt.relation.last UTT 'Token)))                
       (item.set_feat it "punc" "")
       (item.set_feat it "phonemes" (format nil "%l" (list 'pau)))
       (item.set_feat it "dur_stretch" (format nil "%s" msec))
       UTT)))
     (t
       (cond
         ((string-equal "LARGE" level)
           ;; begin new utterance
           (begin
             (xxml_synth UTT)
             nil))
         ((string-equal "NONE" level) 
           (let ((last_token (utt.relation.last UTT 'Token)))
           (if last_token
             (item.set_feat last_token "pbreak" "NB"))
           UTT))
         (t
          (let ((last_token (utt.relation.last UTT 'Token)))
          (if last_token
            (item.set_feat last_token "pbreak" "B")))
          UTT))))))

 ("(DIV" (ATLIST UTT)
   (xxml_synth UTT)
   nil)
  ("AUDIO" (ATTLIST UTT)
    ;; insert "AUDIO" token with descriptor for audiofile
    ;; this must be picked up by after_synth hook
     (let ((src (car (xxml_attval "SRC" ATTLIST)))
           (last_token (utt.relation.last UTT 'Token)))
     (if last_token 
       (item.insert last_token '(AUDIO ((prepunctuation ""))))
       (utt.relation.append UTT 'Token '(AUDIO ((prepunctuation "")))))
     (let ((it (utt.relation.last UTT 'Token)))                
     ;(item.set_feat it "phonemes" (format nil "%l" nil))
     ; preferring to place short pause to not interrupt continuous speech
     (item.set_feat it "phonemes" (format nil "%l" (list 'pau)))
     ; apparently, pause cannot be compressed below 10 ms, so following doesn't quite work
     (item.set_feat it "punc" "")
     (item.set_feat it "dur_stretch" (format nil "%s" 0.0001))
     (item.set_feat it "audiofile" (format nil "%s" src))
     UTT)))    
   ;; JW: the following is Edinburgh's version for the AUDIO command:
   ;; Status: MODE (background) ignored, only insertion supported
   ;; mime type of file also ignored, as its LEVEL
     ;(let ((tmpfile (make_tmp_filename)))
     ;; ignoring mode-background (and will for sometime)
     ;; ignoring level option
     ;(xxml_synth UTT)  ;; synthesizing anything ready to be synthesized
     ;(get_url (car (xxml_attval "SRC" ATTLIST)) tmpfile)
     ;(apply_hooks tts_hooks
     ; (eval (list 'Utterance 'Wave tmpfile)))
     ;(delete-file tmpfile)
     ;nil))
  ("(PITCH" (ATTLIST UTT)
   (ogisable_push_word_features)
   ;(xxml_synth UTT)
   (set! ogisable_pitch_context (cons int_lr_params ogisable_pitch_context))
   (let (
   (base (ogisable_interpret_param
     (car (xxml_attval "BASE" ATTLIST))
     ogisable_pitch_base_map
     ogisable_pitch_base_original
     (cadr (assoc 'target_f0_mean int_lr_params))
     ))
   (tobi (car (xxml_attval "TOBI" ATTLIST)))
   (contour (car (xxml_attval "CONTOUR" ATTLIST)))
   (med (ogisable_interpret_param
         (car (xxml_attval "MED" ATTLIST))
         ogisable_pitch_med_map
         ogisable_pitch_med_original
         (cadr (assoc 'target_f0_mean int_lr_params))
         ))
   (range (ogisable_interpret_param
     (car (xxml_attval "RANGE" ATTLIST))
     ogisable_pitch_range_map
     ogisable_pitch_range_original
     (cadr (assoc 'target_f0_std int_lr_params))
     ))
   (oldmean (cadr (assoc 'target_f0_mean int_lr_params)))
   (fv) (tone) (accent))
     ;; Festival (if it supports anything) supports mean and std
     ;; so we treat base as med if med doesn't seem to do anything
     (if (equal? med oldmean)
       (set! med base))
     ;(set! int_lr_params
      ; (cons
      ; (list 'target_f0_mean med)
      ; (cons
      ;  (list 'target_f0_std range)
      ;  int_lr_params)))
     (if (symbol-bound? 'pitchID)
       (set! pitchID (+ pitchID 1))
       (set! pitchID 1))
     (set! xxml_word_features 
     (cons (list "ogisable_pitchID" pitchID) xxml_word_features))
     (if contour 
       (set! xxml_word_features 
       (cons (list "ogisable_f0contour" (format nil "%l" contour)) xxml_word_features)))
     (if tobi (begin
       (if (string-matches tobi ".*;.*") (begin
         (set! accent (string-before tobi ";"))
         (set! tone (string-after tobi ";")))
         ;; else 
         (begin
         (set! accent tobi)
         (set! tone tobi)))
       (if (or (string-matches accent ".*[LH]\\*") (string-matches accent "NONE")) (begin
         (set! fv (append fv (list (list 'accent accent))))))
       (if (string-matches tone "[LH]-[LH]%") 
         (set! fv (append fv (list (list 'tone tone)))))
       (set! xxml_word_features 
       (append 
       ;(list "ogisable_tobi" (format nil "%l" tobi)) 
       fv
       xxml_word_features))))
     (if med
       (set! xxml_word_features 
       (cons (list "ogisable_f0base" (format nil "%l" med)) xxml_word_features)))
     (if range
       (set! xxml_word_features 
       (cons (list "ogisable_f0range" (format nil "%l" range)) xxml_word_features)))
      
   UTT))
  (")PITCH" (ATTLIST UTT)
   (set! int_lr_params (car ogisable_pitch_context))
   (set! ogisable_pitch_context (cdr ogisable_pitch_context))
   (set! xxml_word_features (ogisable_pop_word_features))
   UTT)
  ("(VOLUME" (ATTLIST UTT)
   (ogisable_push_word_features)
   (if (symbol-bound? 'volumeID)
     (set! volumeID (+ volumeID 1))
     (set! volumeID 1))
   (set! ogisable_vol_context (cons ogisable_vol_factor ogisable_vol_context))
   (let ((level (ogisable_interpret_param
                  (car (xxml_attval "LEVEL" ATTLIST))
                  ogisable_volume_level_map
                  1.0
                  ogisable_vol_factor))
          (contour (car (xxml_attval "CONTOUR" ATTLIST))))
      ; JW: disable absolute volume controle
      ;((string-equal (car (xxml_attval "LEVEL" ATTLIST))  SABLE_RXDOUBLE)
      ; (set! ogisable_vol_type 'absolute))
     (if (string-matches (car (xxml_attval "LEVEL" ATTLIST)) ".*%")
       (set! contour (string-append 0 " " ogisable_vol_factor ";"
                                    0.1 " " level ";" 0.9 " " level ";" 
                                    1 " " ogisable_vol_factor)))
     (if contour (begin
       (set! xxml_word_features 
       (cons (list "ogisable_Vcontour" contour) xxml_word_features))
       (set! xxml_word_features 
       (cons (list "ogisable_volumeID" volumeID) xxml_word_features))
       )))
   UTT)
  (")VOLUME" (ATTLIST UTT)
   (set! xxml_word_features (ogisable_pop_word_features))
   (set! ogisable_vol_factor (car ogisable_vol_context))
   (set! ogisable_vol_context (cdr ogisable_vol_context))
   UTT)
  ("(VQUALITY" (ATTLIST UTT)
   (ogisable_push_word_features)
   (if (symbol-bound? 'vqualityID)
     (set! vqualityID (+ vqualityID 1))
     (set! vqualityID 1))
   (set! contour (car (xxml_attval "CONTOUR" ATTLIST)))
   (set! xxml_word_features 
   (cons (list "ogisable_VQcontour" contour) xxml_word_features))
   (set! xxml_word_features 
   (cons (list "ogisable_vqualityID" vqualityID) xxml_word_features))
   UTT)
  (")VQUALITY" (ATTLIST UTT)
   (set! xxml_word_features (ogisable_pop_word_features))
   UTT)
  ("(ENGINE" (ATTLIST UTT)
   ;; Status: probably complete
   (xxml_synth UTT)
   (if (string-matches (car (xxml_attval "ID" ATTLIST)) "festival.*")
       (let ((datastr ""))
   (mapcar
    (lambda (c) (set! datastr (string-append datastr " " c)))
    (xxml_attval "DATA" ATTLIST))
   (apply_hooks tts_hooks (eval (list 'Utterance 'Text datastr)))
   (set! ogisable_omitted_mode t)) ;; ignore contents 
       ;; else 
       ;;  its not relevant to me
       )
   nil)
  (")ENGINE" (ATTLIST UTT)
   (xxml_synth UTT)
   (set! ogisable_omitted_mode nil)
   nil)

   ;; the following commands do not cause utterance breaks
   ;; instead they return UTT 
  ("(EMPH" (ATTLIST UTT)
   ;; Status: nesting makes no difference, levels ignored
   ;; Festival is particularly bad at adding specific emphasis
   ;; that's what happens when you use statistical methods that
   ;; don't include any notion of emphasis
   ;; This is *not* recursive and only one level of EMPH supported
   (ogisable_push_word_features)
   (let ((level (downcase (car (xxml_attval "LEVEL" ATTLIST)))))
     (cond
       ((string-equal level "strong") (set! level 1.9))
       (t (set! level 1.6)))
     (set! level (* level 
       (ogisable_find_fval "dur_stretch" xxml_word_features 1.0)))
     (set! xxml_word_features 
     (append 
             ;(list (list "EMPH" "1")) 
             (list (list "tone" "H*")) 
             xxml_word_features 
             (list (list "dur_stretch" level))))
     UTT))
  (")EMPH" (ATTLIST UTT)
   (set! xxml_word_features (ogisable_pop_word_features))
   UTT)
  ("(RATE" (ATTLIST UTT)
   ;; Status: can't deal with absolute word per minute SPEED.
   (ogisable_push_word_features)
   ;; can't deal with words per minute value
   (let ((rate (ogisable_interpret_param
    (car (xxml_attval "SPEED" ATTLIST))
    ogisable_rate_speed_map
    (ogisable_find_fval "dur_stretch" xxml_word_features 1.0)
    ogisable_rate_speed_original
     )))
     ; should replace dur_stretch feature rather than add new pair at end of list
     (set! xxml_word_features 
     (append  xxml_word_features (list (list "dur_stretch" (/ 1.0 rate)))))
     UTT))
  (")RATE" (ATTLIST UTT)
   (set! xxml_word_features (ogisable_pop_word_features))
   UTT)
  ("MARKER" (ATTLIST UTT)
    ;; insert "MARK" token and add mark string as feature
    ;; this must be picked up by after_synth hook
     (let ((mark (car (xxml_attval "EMBED" ATTLIST)))
           (last_token (utt.relation.last UTT 'Token)))
     (if last_token 
       (item.insert last_token '(MARK ((prepunctuation ""))))
       (utt.relation.append UTT 'Token '(MARK ((prepunctuation "")))))
     (let ((it (utt.relation.last UTT 'Token)))                
     (item.set_feat it "phonemes" (format nil "%l" nil))
     (item.set_feat it "punc" "")
     (item.set_feat it "mark" (format nil "%s" mark))
     UTT)))    
   ;; JW: below is Edinburgh version for MARKER
   ;; Status: does nothing
   ;; Can't support this without low-level control of audio spooler
   ;;(format t "SABLE: marker \"%s\"\n" 
   ;;  (car (xxml_attval "MARK" ATTLIST))
   ;;UTT)
  ("(PRON" (ATTLIST UTT)
   (ogisable_push_word_features)
   (if (symbol-bound? 'pronID)
     (set! pronID (+ pronID 1))
     (set! pronID 1))
   ;; can't deal with words per minute value
   (let ((ipa (xxml_attval "IPA" ATTLIST))
         (sub (xxml_attval "SUB" ATTLIST))
         (segdur (car (xxml_attval "SEGDUR" ATTLIST))))
     (cond
      (ipa
       (set! xxml_word_features
       (cons (list "ogisable_ipa" (format nil "%l" ipa)) xxml_word_features))
       (set! xxml_word_features
       (cons (list "ogisable_pronID" pronID) xxml_word_features))
       (set! xxml_word_features 
       (cons (list "ogisable_ignore" "1") xxml_word_features)))
       ;; previous version with ipa in text not in tag :
       ;;(cons (list "ogisable_ipa" "1") xxml_word_features)
      (segdur
       (set! xxml_word_features
       (cons (list "ogisable_segdur" (format nil "%l" segdur)) xxml_word_features))
       (set! xxml_word_features
       (cons (list "ogisable_pronID" pronID) xxml_word_features))
       (set! xxml_word_features 
       (cons (list "ogisable_ignore" "1") xxml_word_features)))
      (sub
       (set! xxml_word_features 
       (cons (list "ogisable_sub" (format nil "%l" sub)) xxml_word_features))
       (set! xxml_word_features
       (cons (list "ogisable_pronID" pronID) xxml_word_features))
       (set! xxml_word_features 
       (cons (list "ogisable_ignore" "1") xxml_word_features))))
     UTT))
  (")PRON" (ATTLIST UTT)
   (set! xxml_word_features (ogisable_pop_word_features))
   UTT)
  ("(SAYAS" (ATTLIST UTT)
   ;; Status: only a few of the types are dealt with
   (ogisable_push_word_features)
    (set! ogisable_utt UTT)
   ;; can't deal with words per minute value
   (let ((mode (downcase (car (xxml_attval "MODE" ATTLIST))))
         (modetype (car (xxml_attval "MODETYPE" ATTLIST))))
     (cond
      ((string-equal mode "literal")
       (set! xxml_word_features 
       (cons (list "ogisable_literal" "1") 
         (cons (list "spell_mode" (format nil "%s" (downcase modetype))) xxml_word_features))))
         ;;xxml_word_features
      ((string-equal mode "syllabify")
       (set! xxml_word_features 
       (cons (list "ogisable_syllabify" "1") 
         (cons (list "syllabify_info" (format nil "%s" modetype)) xxml_word_features))))
      ((string-equal mode "phone")
       (set! xxml_word_features 
       (cons (list "token_pos" "digits") xxml_word_features)))
      ((string-equal mode "ordinal")
       (set! xxml_word_features 
       (cons (list "token_pos" "ordinal") xxml_word_features)))
      ((string-equal mode "cardinal")
       (set! xxml_word_features 
       (cons (list "token_pos" "cardinal") xxml_word_features)))
      (t
       ;; blindly trust festival to get it right 
       t))
     UTT))
  (")SAYAS" (ATTLIST UTT)
   (set! xxml_word_features (ogisable_pop_word_features))
   UTT)
))

(define (ogisable_init_func)
  "(ogisable_init_func)
Initialisation for SABLE mode"
  (set! counter 0)
  (ogisable_init_globals)
  (set! sounds_directory (path-append libdir 'sounds))
  (set! ogisable_original_token_to_words token_to_words)
  (set! ogisable_original_Intonation Intonation)
  (set! ogisable_previous_elements xxml_elements)
  (set! xxml_elements ogisable_elements)
  (set! ogisable_current_speaker 'voice_mwm_diphone)
  (init_ogisable_token_to_words)
  (if (and after_synth_hooks (not (consp after_synth_hooks))) 
    (set! after_synth_hooks (list after_synth_hooks)))
  (set! previous_after_synth_hooks after_synth_hooks)
  (set! previous_after_analysis_hooks after_analysis_hooks)
  (set! previous_pos_map pos_map)
  ;; this is reset when voice changes (why?) so repeat also in setup_voice_params:
  (set! after_synth_hooks (append after_synth_hooks (list ogisable_insert_audio)))
  (set! pos_map nil)
  (set! after_analysis_hooks 
    (append (list ogisable_adjust_segments ogisable_apply_segdurs)
            (list ogisable_apply_f0contour ogisable_adjust_volume ogisable_adjust_vquality)
            after_analysis_hooks))
  )
  
(define (save_waveform utt)
  (print "saving waveform")
  (utt.save.wave utt '/temp/tmp.wav 'riff))

(define (init_ogisable_token_to_words)
   "(init_ogisable_token_to_words)
Hook in markup token-to-word rules."
  ;; token_to_words can be modified by voice definition 
  (if (not (equal? ogisable_token_to_words token_to_words))
    (set! ogisable_previous_token_to_words token_to_words) nil)
  (if (not (equal? ogisable_Intonation Intonation))
    (set! ogisable_previous_Intonation Intonation) nil)
  (set! token_to_words ogisable_token_to_words)
  (set! Intonation ogisable_Intonation))
  

(define (ogisable_exit_func)
  "(ogisable_exit_func)
Exit function for SABLE mode"
  (set! after_synth_hooks previous_after_synth_hooks)   
  (set! after_analysis_hooks previous_after_analysis_hooks)   
  (set! xxml_elements ogisable_previous_elements)
  (set! pos_map previous_pos_map)
  (set! Intonation ogisable_original_Intonation)
  (set! token_to_words ogisable_original_token_to_words))
  
(define (ogisable_push_word_features)
"(ogisable_push_word_features)
Save current word features on stack."
  (set! ogisable_word_features_stack 
  (cons xxml_word_features ogisable_word_features_stack)))

(define (ogisable_insert_audio utt)
  "(ogisable_insert_audio utt)
Insert audiofiles into synthesized waveform, and adjust utterance durations.
The descriptors for the audiofiles are in the token feature *audiofile*."
  ;; find "audiofile" features
  ;(print (utt.features utt 'IntEvent '(name R:Intonation.parent.R:SylStructure.parent.name)))
  ;(print (utt.features utt 'Segment '(name end)))
  (let ((token (utt.relation.first utt 'Token))
        (audiofile nil) (time 0) (file nil) (dur 0))
  (while token (begin
    (set! audiofile (item.feat token 'audiofile))
    (if (not (equal? audiofile 0)) (begin
      (set! time (item.feat token 'daughter1.R:SylStructure.daughter1.daughter1.segment_start))     
      ;(if (equal? (item.feat token 'p.audiofile) 0)
      ;  (set! time (item.feat token 'p.daughter1.R:SylStructure.daughtern.daughtern.end))
        ;; audio segment is not linked properly to word and SylStructure
        ;; in case sequence of audio segments, just inherit time info from previous loop
      ; (set! time (+ time dur)))
      ;(if (<= time 0) (set! time 0.001) nil) ;; bug in insert if time==0
      ;; if audiofile is URL, then create tmp file
      (if (string-matches audiofile "http://.*") 
        (begin
          (let ((tmpfile (make_tmp_filename)))
          (get_url audiofile tmpfile)     
          (utt.wave.overwrite utt tmpfile time) 
          (delete-file tmpfile)))
        (begin
          (set! file (path-append sounds_directory audiofile))
          (cond 
            ((probe_file audiofile) (utt.wave.overwrite utt audiofile time))  
            ((probe_file file) (utt.wave.overwrite utt file time))
            (t (format t "could not find %s or %s\n" audiofile file)))
          ))))
     (set! token (item.next token))))
   ;(utt.save.wave utt (format nil "tmp%s.wav" counter)) 
   ;(print (utt.features utt 'Segment '(name end audiofile mark)))
   (set! counter (+ counter 1))
   ;(save_waveform utt)
   utt
   ))

(define (ogisable_apply_segdurs utt)
  "(ogisable_apply_segdurs utt)
Apply segment durations. The duration info is retrieved from the durations feature on 
token items."
  (let ((token (utt.relation.first utt 'Token)) (durs) (seg1) (segn) (seg) 
        (oldend) (newend) (diff))
  (while token
    (set! durs (item.feat token "durations"))
    (if (not (equal? durs 0)) (begin
      (set! durs (read-from-string durs)) 
      (if (< 0 (length durs)) (begin
        ;; for each segment in this token, overwrite duration.
        ;; in general it's not safe to assume the following gives the first and last segments
        ;; for token, because certain words have a nil pronunciation. However, for segdurs the
        ;; pronunciation has been explicitly specified using phonetic symbols.
        (set! seg1 (item.relation (item.daughter1 (item.daughter1 
                   (item.relation (item.daughter1 token) 'SylStructure))) 'Segment))
        (set! segn (item.relation (item.daughtern (item.daughtern 
                   (item.relation (item.daughtern token) 'SylStructure))) 'Segment))
        (set! segn (item.next segn))
        (set! seg seg1)
        (while (not (equal? seg segn))
          (if (not durs) 
            (print "ogisable_apply_segdurs : not enough segment durations found for token")
            ;; else
            (begin 
            (set! oldend (item.feat seg "segment_end"))
            (set! newend (+ (item.feat seg "segment_start") (/ (car durs) 1000)))
            (item.set_feat seg "end" newend)
            (set! durs (cdr durs))))
          (set! seg (item.next seg))) ;; builds on Segment relation being a flat list
        ;; fix end times of all later segments
        (set! diff (- newend oldend))
        (while seg
          (item.set_feat seg "end" (+ (item.feat seg "end") diff))
          (set! seg (item.next seg)))))))
    (set! token (item.next token)))
  ;(print (utt.features utt 'Segment '(name end)))
  utt))
   
   
   
(define (ogisable_apply_f0contour utt)
  "(ogisable_apply_f0contour utt)
Apply pitch contour. The contour information is retrieved from the f0contour 
feature on token items."
  (let ((token (utt.relation.first utt 'Token)) (contour) (bt 0) (et 0) (dur) (lasttoken)
        (targets (utt.relation.leafs utt 'Target)) (target nil) (x) (seg) 
        (segend) (seg1) (segn) (coord) (pseg) (nseg)
        (pitchID) (base) (range) (oldbase) (oldrange) (rangeR))
    (while token
      ;; find block of tokens for which contour has been specified
      (set! pitchID (item.feat token 'ogisable_pitchID))
      (set! contour (item.feat token 'ogisable_f0contour))
      (set! base (item.feat token 'ogisable_f0base))
      (set! range (item.feat token 'ogisable_f0range))
      (if (and (or (not (equal? 0 contour))
                   (not (equal? 0 base))
                   (not (equal? 0 range))) 
               (not (string-equal pitchID (item.feat token 'p.ogisable_pitchID)))) (begin
        ;; find begin and end segments of block
        (set! segs (utt.relation.items utt 'Segment))
        (while (and segs (not (string-equal 
               (item.feat (car segs) 'R:SylStructure.parent.parent.R:Token.parent.ogisable_pitchID) pitchID)))
          (set! segs (cdr segs)))
        (set! seg1 (car segs))
        (set! segs (cdr segs))
        (while (and segs (string-equal 
               (item.feat (car segs) 'R:SylStructure.parent.parent.R:Token.parent.ogisable_pitchID) pitchID)) 
          (set! segs (cdr segs)))
        (if segs (set! segn (car segs)) (set! segn nil))
        (if seg1 (set! bt (item.feat seg1 "segment_start")) 
                 (print "ogisable_apply_f0contour: no first segment in tag"))
        (if segn (set! et (item.feat segn "segment_start"))
                 (print "ogisable_apply_f0contour: no last segment in tag"))
        (if (and seg1 segn) (begin
          (set! dur (- et bt))
          ;; replace f0 targets for each segment between bt and et by specified contour 
          (if (not (equal? 0 contour)) (begin 
            ;; delete old targets
            (set! seg seg1) 
            (while (not (equal? seg segn))
              (if (item.relation seg 'Target) (begin
                (item.relation.remove (item.relation seg 'Target) 'Target)
                ;(format t "deleted segment %s in Target stream\n" (item.feat seg "name"))
                ))
              (set! seg (item.next seg)))
            ;; create new targets
            (set! seg seg1)
            (set! contour (parse-contour contour))
            (while contour 
              (set! coord (car contour))
              (set! x (+ bt (* (car coord) dur)))
              (if seg (set! segend (item.feat seg "end")))
              (while (and (not (equal? seg segn)) (< segend x))
                (set! seg (item.next seg))
                (if seg (set! segend (item.feat seg "end"))))
              (if (or (equal? seg segn) (< x (item.feat seg "segment_start")))
                (format t "ogisable_apply_f0contour: segment end times not conform target at %f\n" x)
                ;; else
                (begin 
                ;; insert segment in Target relation if necessary
                (set! Tseg (item.relation seg 'Target))
                (if (not Tseg) (begin
                   ;; find neighbor in Target relation
                   (set! pseg (item.prev seg))
                   (while (and pseg (not (item.relation pseg 'Target))) 
                     (set! pseg (item.prev pseg)))
                   (if pseg 
                     (set! Tseg (item.insert (item.relation pseg 'Target) seg))
                     ;; else
                     (begin
                       (set! nseg (item.next seg))
                       (while (and nseg (not (item.relation nseg 'Target))) 
                         (set! nseg (item.next nseg)))
                       (if nseg 
                         (set! Tseg (item.insert (item.relation nseg 'Target) seg 'before))
                         ;; else (Target relation must be empty) 
                         (begin
                         ;(print "ogisable_apply_f0contour: didn't expect empty Target relation")
                         (utt.relation.create utt 'Target)
                         (set! Tseg (utt.relation.append utt 'Target seg)))
                         )))))                   
                (if Tseg (begin 
                  (set! target (item.append_daughter Tseg "0"))
                  (item.set_feat target "pos" x)
                  (item.set_feat target "f0" (cadr coord)))
                  ;;else
                  (print "ogisable_apply_f0contour: segment could not be created in Target relation"))))
              (set! contour (cdr contour))))
          ;; else apply base and range
            (begin
            (set! seg seg1)
            (set! oldbase (cadr (assoc 'target_f0_mean int_lr_params)))
            (set! oldrange (cadr (assoc 'target_f0_std int_lr_params)))
            (if (equal? 0 base) (set! base oldbase) (set! base (parse-number base)))
            (if (equal? 0 range) (set! range oldrange) (set! range (parse-number range)))
            (set! rangeR (/ range oldrange))
            (while (not (equal? seg segn))
              (if (item.relation seg 'Target) (begin
                (set! target (item.daughter1 (item.relation seg 'Target)))
                (while target
                  (item.set_feat target "f0" (+ base (* rangeR (- (item.feat target "f0") oldbase))))
                  (item.feat target "f0")
                  (set! target (item.next target)))))
                 (set! seg (item.next seg)))))))))
      (set! token (item.next token)))
      ;(print (utt.features utt 'Syllable '(R:SylStructure.parent.name R:SylStructure.parent.R:Token.parent.accent R:SylStructure.parent.R:Token.parent.tone R:Intonation.daughter1.name R:Intonation.daughter2.name)))
      ;(mapcar (lambda (a) (print (format nil "%s %s" (item.feat a '(item.feat a 'name))) (utt.relation.items utt 'IntEvent))
      utt))
      
(define (ogisable_adjust_segments utt)
  "(ogisable_adjust_segments utt)
Adjust utterance durations so that audiofile can take place of 'pau'. 
Similarly, attach MARK info as feature on first segment after MARK"
  ;; find "audiofile" features
  ;(print (utt.features utt 'Token '(name daughter1.R:SylStructure.daughter1.daughtern.name)))
  (let ((token (utt.relation.first utt 'Token))
        (audiofile nil) (time 0) (pau_dur 0) (file nil) (dur nil) (tmputt nil) (mark))
  (while token (begin
    (set! audiofile (item.feat token 'audiofile))
    (if (not (equal? audiofile 0)) (begin
      (set! time (item.feat token 'daughter1.R:SylStructure.daughter1.daughter1.segment_start))
      (set! pau_dur (item.feat token 'daughter1.R:SylStructure.daughter1.daughter1.segment_duration))
      ;; if audiofile is URL, then create tmp file
      (if (string-matches audiofile "http://.*") 
        (begin
          (let ((tmpfile (make_tmp_filename)))
          (get_url audiofile tmpfile)     
          (set! dur (/ 
               (utt.wave.info (set! tmputt (utt.import.wave (Utterance) file)) 'num_samples)
               (utt.wave.info tmputt 'sample_rate)))
          (shift-segments utt time (- dur pau_dur) "audiofile" audiofile)
          (delete-file tmpfile)))
        (begin
          (set! file (path-append sounds_directory audiofile))
          (cond 
            ((probe_file audiofile) (set! file audiofile))  
            ((probe_file file) nil)
            (t (set! dur 0)))
          (if (not dur) 
            (set! dur (/ 
               (utt.wave.info (set! tmputt (utt.import.wave (Utterance) file)) 'num_samples)
               (utt.wave.info tmputt 'sample_rate))))
          (shift-segments utt time (- dur pau_dur) "audiofile" file)
       ))))
       ;(item.set_feat token "time" (format nil "%s" time))
    ;; find marker, attach feature value pair to first segment after mark
    (set! mark (item.feat token 'mark))
    (if (not (equal? mark 0)) (begin
      (if (equal? (item.feat token 'p.mark) 0)
        ;; find end of last segment (will find 0 if no preceding segments)
        (set! time (item.feat token 'p.daughter1.R:SylStructure.daughtern.daughtern.end))
        ;; in case sequence of marks, just inherit time info from previous loop
        (set! time time))
      (shift-segments utt time 0 "mark" mark)
      ))
      ;(item.set_feat token "time" (format nil "%s" time))
    (set! token (item.next token))))
   utt
   ))

(define (shift-segments utt time dur feature value)
  "(shift-segments utt time dur feature value)
Stretch first segment after *time* with *dur* seconds, and 
shift end times of later segments. Add feature value pair on stretched segment."
  (let ((seg (utt.relation.first utt 'Segment)) (end nil) 
        (firstseg 'true) (targs (utt.relation.leafs utt 'Target)) (targ))
  (while seg (begin
     (set! end (item.feat seg 'end))
     (if (> end time) (begin
       (if firstseg (begin
         ; first segment after 'time': stretch and attach feature value pair
         (set! firstseg nil)
         (item.set_feat seg "end" (+ end dur))
         (if feature (item.set_feat seg feature value)))
         ; else
         (item.set_feat seg "end" (+ end dur))
         )))
     (set! seg (item.next seg))))
  (while targs (begin
     (set! targ (car targs))
     (set! end (item.feat targ "pos"))
     (if (> end time)
       (item.set_feat targ "pos" (+ end dur)))
     (set! targs (cdr targs))))
     ))

(define (ogisable_adjust_volume utt)
  "(ogisable_adjust_volume utt)
Amplify or attenutate signal based on contour in token.ogisable_Vcontour"
  (let ((token (utt.relation.first utt 'Token)) (volumeID) (contour)
        (segs) (seg1) (segn) (bt) (et) (x) (dur))
  (utt.relation.create utt 'Gain)
  (while token (begin
    (set! volumeID (item.feat token 'ogisable_volumeID))
    (set! contour (item.feat token "ogisable_Vcontour"))
    (if (and (not (equal? 0 contour))
             (not (string-equal volumeID (item.feat token 'p.ogisable_volumeID)))) (begin
      ;; find duration of contour
      (set! segs (utt.relation.items utt 'Segment))
      (while (and segs (not (string-equal 
             (item.feat (car segs) 'R:SylStructure.parent.parent.R:Token.parent.ogisable_volumeID) volumeID)))
        (set! segs (cdr segs)))
      (set! seg1 (car segs))
      (set! segs (cdr segs))
      (while (and segs (string-equal 
             (item.feat (car segs) 'R:SylStructure.parent.parent.R:Token.parent.ogisable_volumeID) volumeID)) 
        (set! segs (cdr segs)))
      (if segs (set! segn (car segs)) (set! segn nil))
      (if (and seg1 segn) (begin
        (set! bt (item.feat seg1 "segment_start"))        
        (set! et (item.feat segn "segment_start"))
        (set! dur (- et bt))
        (set! contour (parse-contour contour))
        (while contour 
          (set! coord (car contour))
          (set! x (+ bt (* (car coord) dur)))
          (utt.relation.append utt 'Gain 
            (list volumeID (list (list "pos" x) (list "gain" (cadr coord)))))
          (set! contour (cdr contour))    
          )))))
      (set! token (item.next token))))
      ;(print (utt.features utt 'Gain '(name pos gain)))
      ;(utt.wave.OGIrescale utt)
      utt))
        
(define (ogisable_adjust_vquality utt)
  "(ogisable_adjust_vquality utt)
Amplify or attenutate signal based on contour in token.ogisable_VQcontour"
  (let ((token (utt.relation.first utt 'Token)) (vqualityID) (contour)
        (segs) (seg1) (segn) (bt) (et) (x) (dur))
  (while token (begin
    (set! vqualityID (item.feat token 'ogisable_vqualityID))
    (set! contour (item.feat token "ogisable_VQcontour"))
    (if (and (not (equal? 0 contour))
             (not (string-equal vqualityID (item.feat token 'p.ogisable_vqualityID)))) (begin
      (if (not (utt.relation.present utt 'VQuality))
        (utt.relation.create utt 'VQuality))
      ;; find duration of contour
      (set! segs (utt.relation.items utt 'Segment))
      (while (and segs (not (string-equal 
             (item.feat (car segs) 'R:SylStructure.parent.parent.R:Token.parent.ogisable_vqualityID) vqualityID)))
        (set! segs (cdr segs)))
      (set! seg1 (car segs))
      (set! segs (cdr segs))
      (while (and segs (string-equal 
             (item.feat (car segs) 'R:SylStructure.parent.parent.R:Token.parent.ogisable_vqualityID) vqualityID)) 
        (set! segs (cdr segs)))
      (if segs (set! segn (car segs)) (set! segn nil))
      (if (and seg1 segn) (begin
        (set! bt (item.feat seg1 "segment_start"))        
        (set! et (item.feat segn "segment_start"))
        (set! dur (- et bt))
        (set! contour (parse-contour contour))
        (while contour 
          (set! coord (car contour))
          (set! x (+ bt (* (car coord) dur)))
          (utt.relation.append utt 'VQuality 
            (list vqualityID (list (list "pos" x) (list "vquality" (cadr coord)))))
          (set! contour (cdr contour))    
          )))))
      (set! token (item.next token))))
      ;(print (utt.features utt 'VQuality '(name pos vquality)))
      utt))
        
        
(define (ogisable_adjust_volume_old utt)
  "(ogisable_adjust_volume utt)
Amplify or attenutate signale based on value of ogisable_vol_factor
and ogisable_vol_type (absolute or relative)."
  (set! utts (cons utt utts))
  (cond
   ((equal? ogisable_vol_type 'no_change)
    utt)
   ((equal? ogisable_vol_type 'absolute)
    (utt.wave.rescale utt ogisable_vol_factor 'absolute))
   ((equal? ogisable_vol_type 'relative)
    (utt.wave.rescale utt ogisable_vol_factor))
   (t
    (format stderr "SABLE: volume unknown type \"%s\"\n" ogisable_vol_type)
    utt))
   utt)

(define (ogisable_pop_word_features)
"(ogisable_pop_word_features)
Pop word features from stack."
  (let ((r (car ogisable_word_features_stack)))
    (set! ogisable_word_features_stack (cdr ogisable_word_features_stack))
    r))

(define (ogisable_find_fval feat flist def)
  (cond
   ((null flist) def)
   ((string-equal feat (car (car flist)))
    (car (cdr (car flist))))
   (t
    (ogisable_find_fval feat (cdr flist) def))))

(define (ogisable_interpret_param ident map original current)
"(ogisable_interpret_param IDENT MAP ORIGINAL CURRENT)
If IDENTity is in MAP return ORIGINAL times value in map, otherwise
treat IDENT of the form +/-N% and modify CURRENT accordingly."
  (let ((mm (assoc ident map)))
    (cond 
     (mm
      (* original (car (cdr mm))))
     ((string-equal ident SABLE_RXDOUBLE)
      (parse-number ident))
     ((string-matches ident ".*%")
      (+ current (* current (/ (parse-number (string-before ident "%")) 
             100.0))))
;;     ((string-matches ident ".*%")
;;      (* current (/ (parse-number (string-before ident "%")) 100.0)))
     ((not ident) current)
     (t
      (format stderr "SABLE: modifier \"%s\" not of float, tag or +/-N\n"
        ident)
      current))))

(define (ogisable_setup_voice_params)
"(ogisable_setup_voice_params)
Set up original values for various voice parameters."
 ;; almost seems that we need special new-voice command for markup
 ;; when voice changes, the following may have to be reset:
 ;; token_to_words, voice_params, (lexicon rules?)
 (set! ogisable_pitch_base_original (cadr (assoc 'target_f0_mean int_lr_params)))
 (set! ogisable_pitch_med_original (cadr (assoc 'target_f0_mean int_lr_params)))
 (set! ogisable_pitch_range_original (cadr (assoc 'target_f0_std int_lr_params)))
 (set! ogisable_rate_speed_original 1.0)
 (set! pos_map nil)
 (if (not (member ogisable_insert_audio after_synth_hooks))
   (set! after_synth_hooks (append previous_after_synth_hooks (list ogisable_insert_audio))) nil)
 ;(if (and after_synth_hooks (not (consp after_synth_hooks)))
 ;    (set! after_synth_hooks 
 ;    (cons after_synth_hooks (list ogisable_adjust_volume 
 ;    )))
 ;    (set! after_synth_hooks 
 ;    (append after_synth_hooks (list ogisable_adjust_volume   
 ;    ))))     
)

(define (ogisable-map-segdur symbols)
"(ogisable-map-segdur symbols)
Map input string of form 'pau 20; A 100; 9r 80; ...' to ((pau A 9r ..) (20 100 80 ..))"
  (let ((segs nil) (durs nil) (tuple) (newsegdur))
    (while (< 0 (string-length symbols))
      (if (string-matches symbols ".*;.*")
        (set! tuple (string-before symbols ";"))
        (set! tuple symbols))
      (set! symbols (string-after symbols ";"))
      (set! newsegdur (read-from-string (string-append "(" tuple ")")))
      (if (phone_feature (car newsegdur) 'cvox) (begin
        (set! segs (append segs (list (car newsegdur))))
        (set! durs (append durs (list (cadr newsegdur)))))))
  (list segs durs)))
(ogisable-map-segdur "pau 20; A 100; 9r 209")       

(define (ogisable-map-ipa symbols)
"(ogisable-map-ipa symbols)
Map phonetic symbols to worldbet symbols. At the moment expected 
input is worldbet, but could be other symbols could also be supported.
For unknown symbols the empty string is returned."
(let ((safesymbols nil) (symbol (car symbols)))  
  (while symbol (begin
    (if (or (phone_feature (string-before symbol '1) 'cvox)
        (phone_feature symbol 'cvox))
        ;; cheap way to see if symbol defined in current phoneset
      (set! safesymbols (append safesymbols (list symbol))))
    (set! symbols (cdr symbols))
    (set! symbol (car symbols))))
  safesymbols 
  ))

(define (make-syl-list syll_info)
"(make-syl-list syll_info)
Convert string of phonetic symbols and . into list of syllables, where each syllable
is a list of phonetic symbols that are in the phoneset."
(let ((symbol (substring syll_info 0 1)) (phon "") (phons) (syls))
  ;; remove initial spaces or ., and attach . at end
 (while (or (string-equal symbol " ") (string-equal symbol ".")) (begin
    (set! syll_info (substring syll_info 1 (string-length syll_info)))
    (set! symbol (substring syll_info 0 1))))
  (set! syll_info (string-append syll_info "."))
  (while (not (string-equal syll_info "")) (begin
    ;; find next phoneme
    (set! symbol (substring syll_info 0 1))
    (if (not (or (string-equal symbol " ") (string-equal symbol ".")))
      (set! phon (string-append phon symbol))
      ;; else
      (begin
        (set! phons (append phons (list phon)))
        (set! phon "")))
    (if (string-equal symbol ".") (begin      
      (set! syls (append syls (list (ogisable-map-ipa phons))))
      (set! phons nil)))
    (set! syll_info (substring syll_info 1 (string-length syll_info)))))
syls))

(define (syllabify phons)
"(syllabify phons)
Convert list of phonetic symbols into list of syllables, using the following heuristics:
V.CV, VC.CV, VC.CCV, VCC.CCV, VCC.CCCV where V=vowel, C=consonant, .=syllable boundary"
(let ((phon (car phons)) (cons) (syl) (syls) (splitcons))
 ;; append all initial consonants and first vowel to first syl
 (while (and phons (string-equal (phone_feature phon 'vc) "-")) (begin
   (set! syl (append syl (list (car phons))))
   (set! phons (cdr phons))
   (set! phon (car phons))))
 (if (string-equal (phone_feature phon 'vc) "+") (begin
   (set! syl (append syl (list phon)))
   (set! phons (cdr phons))))
 ;; append consonants to cons, and when vowel comes, split cons and add syllable
 (while phons (begin
   (set! phon (car phons)) 
   (if (string-equal (phone_feature phon 'vc) "+") (begin
     (set! splitcons (split cons))
     (set! syl (append syl (car splitcons)))
     (set! cons nil)
     (set! syls (append syls (list syl)))
     (set! syl (append (cadr splitcons) (list phon))))     
     ;; else
     (if (string-equal (phone_feature phon 'vc) "-")
       (set! cons (append cons (list phon)))))
    (set! phons (cdr phons))))
  ;; assign final cons to last syllable and add last syllable
  (set! syl (append syl cons))
  (set! syls (append syls (list syl)))))
  
(define (split cons)
  (let ((len) (splitcons))
    (set! len (length cons))
    (cond
      ((equal? len 0)
        (set! splitcons (list nil nil)))
      ((equal? len 1)
        (set! splitcons (list nil cons)))
      ((equal? len 2)
        (set! splitcons (list (list (car cons)) (cdr cons))))
      ((equal? len 3)
        (set! splitcons (list (list (car cons)) (cdr cons))))
      ((equal? len 4)
        (set! splitcons (list (list (car cons) (cadr cons)) (cddr cons))))
      (t
        (set! splitcons (list (list (car cons) (cadr cons)) (cddr cons)))))))

(define (parse-contour contour)
  "(parse-contour contour)
Read list of coordinates from string and return as list. The input format is 
x1 y1 ; x2 y2 ; ... and the output is ((x1 y1) (x2 y2) ..). x1 is between 0 and 1."
  (let ((coord nil) (newcoord) (x) (y) (tuple))
    (while (< 0 (string-length contour))
      (if (string-matches contour ".*;.*")
        (set! tuple (string-before contour ";"))
        (set! tuple contour))
      (set! contour (string-after contour ";"))
      (set! newcoord (read-from-string (string-append "(" tuple ")")))
      (set! x (parse-number (car newcoord)))
      (set! y (parse-number (cadr newcoord)))
      (if (or (< x 0) (> x 1)) 
        (format 't "ogisable_apply_f0contour: x must be between 0 and 1 : %f\n" x)
        (if (> 2 (length newcoord))
          (format 't "ogisable_apply_f0contour: incorrect element in f0contour\n")
          (set! coord (append coord (list (list x y)))))))
    coord))

(define (find_phonemes subword pos)
  (let ((pron) (phons))
    (set! pron (car (cddr (lex.lookup subword pos))))
    (while pron (begin
      ;; only keeping phonemes at the moment, maybe should get stress too
      (set! phons (append phons (caar pron)))
      (set! pron (cdr pron))))
    phons))
    
(define (substring str b n)
  (let ((lstr (symbolexplode str)) (substr "") (i 0)) 
    (while (and lstr (< i b)) (set! i (+ i 1)) (set! lstr (cdr lstr)))
    (set! i 0)
    (while (and lstr (< i n)) 
      (set! i (+ i 1))
      (set! substr (string-append substr (car lstr)))
      (set! lstr (cdr lstr)))
  substr))
   
(define (utt.wave.info u key)
  (let ((fv (wave.info (utt.wave u))))
    (cadr (assoc key fv))))
    
;;; Declare the new mode to Festival
(set! tts_text_modes
   (cons
    (list
      'ogisable   ;; mode name
      (list         
       (list 'init_func ogisable_init_func)
       (list 'exit_func ogisable_exit_func)
       '(analysis_type xml)
       ))
    tts_text_modes))

(provide 'ogisable-mode)
